use super::{
    VerbContext, UserVerb, UserVerbRef, UResult, UserError, user_error,
    get_player_item_or_fail,
    look,
};
use async_trait::async_trait;
use crate::{
    DResult,
    language,
    regular_tasks::queued_command::{
        QueueCommandHandler,
        QueueCommand,
        queue_command
    },
    static_content::room::{self, Direction, ExitType},
    models::item::{
        Item,
        SkillType,
        LocationActionType
    },
    services::{
        broadcast_to_room,
        skills::skill_check_and_grind,
        combat::stop_attacking_mut,
        combat::handle_resurrect,
    }
};
use mockall_double::double;
#[double] use crate::db::DBTrans;
use std::time;

pub async fn announce_move(trans: &DBTrans, character: &Item, leaving: &Item, arriving: &Item) -> DResult<()> {
    let msg_leaving_exp = format!("{} departs towards {}\n",
                                  &character.display_for_sentence(true, 1, true),
                                  &arriving.display);
    let msg_leaving_nonexp = format!("{} departs towards {}\n",
                                     character.display_for_sentence(true, 1, false),
                                     arriving.display_less_explicit
                                       .as_ref()
                                       .unwrap_or(&arriving.display));
    broadcast_to_room(trans, &format!("{}/{}", &leaving.item_type, &leaving.item_code),
                      None, &msg_leaving_exp, Some(&msg_leaving_nonexp)).await?;
    
    let msg_arriving_exp = format!("{} arrives from {}\n", &character.display_for_sentence(true, 1, true),
                                   &leaving.display);
    let msg_arriving_nonexp = format!("{} arrives from {}\n",
                                     character.display_for_sentence(true, 1, false),
                                     leaving.display_less_explicit
                                       .as_ref()
                                       .unwrap_or(&leaving.display));
    broadcast_to_room(trans, &format!("{}/{}", &arriving.item_type, &arriving.item_code),
                      None, &msg_arriving_exp, Some(&msg_arriving_nonexp)).await?;
    Ok(())
}

pub async fn attempt_move_immediate(
    trans: &DBTrans,
    orig_mover: &Item,
    direction: &Direction,
    mut player_ctx: Option<&mut VerbContext<'_>>
) -> UResult<()> {
    let use_location = if orig_mover.is_dead {
        if orig_mover.item_type != "player" {
            user_error("Dead players don't move".to_owned())?;
        }
        "room/repro_xv_respawn"
    } else {
        &orig_mover.location
    };
    let (heretype, herecode) = use_location.split_once("/").unwrap_or(("room", "repro_xv_chargen"));
    if heretype != "room" {
        // Fix this when we have planes / boats / roomkits.
        user_error("Navigating outside rooms not yet supported.".to_owned())?
    }
    let room = room::room_map_by_code().get(herecode)
        .ok_or_else(|| UserError("Can't find your current location".to_owned()))?;
    let exit = room.exits.iter().find(|ex| ex.direction == *direction)
        .ok_or_else(|| UserError("There is nothing in that direction".to_owned()))?;

    let mut mover = (*orig_mover).clone();
    match exit.exit_type {
        ExitType::Free => {}
        ExitType::Blocked(blocker) => {
            if let Some(ctx) = player_ctx.as_mut() {
                if !blocker.attempt_exit(ctx, &mut mover, exit).await? {
                    user_error("Stopping movement".to_owned())?;
                }
            }
        }
    }

    let new_room =
        room::resolve_exit(room, exit).ok_or_else(|| UserError("Can't find that room".to_owned()))?;

    match mover.active_combat.as_ref().and_then(|ac| ac.attacking.clone()) {
        None => {}
        Some(old_victim) => {
            if let Some((vcode, vtype)) = old_victim.split_once("/") {
                if let Some(vitem) = trans.find_item_by_type_code(vcode, vtype).await? {
                    let mut vitem_mut = (*vitem).clone();
                    stop_attacking_mut(trans, &mut mover, &mut vitem_mut, false).await?;
                    trans.save_item_model(&vitem_mut).await?
                }
            }
        }
    }
    match mover.active_combat.clone().as_ref().map(|ac| &ac.attacked_by[..]) {
        None | Some([]) => {}
        Some(attackers) => {
            let mut attacker_names = Vec::new();
            let mut attacker_items = Vec::new();
            if let Some(ctx) = player_ctx.as_ref() {
                for attacker in &attackers[..] {
                    if let Some((acode, atype)) = attacker.split_once("/") {
                        if let Some(aitem) = trans.find_item_by_type_code(acode, atype).await? {
                            attacker_names.push(aitem.display_for_session(ctx.session_dat));
                            attacker_items.push(aitem);
                        }
                    }
                }
            }
            let attacker_names_ref = attacker_names.iter().map(|n| n.as_str()).collect::<Vec<&str>>();
            let attacker_names_str = language::join_words(&attacker_names_ref[..]);
            if skill_check_and_grind(trans, &mut mover, &SkillType::Dodge, attackers.len() as f64 + 8.0).await? >= 0.0 {
                if let Some(ctx) = player_ctx.as_ref() {
                    trans.queue_for_session(ctx.session,
                                            Some(&format!("You successfully get away from {}\n",
                                                          &attacker_names_str))).await?;
                }
                for item in &attacker_items[..] {
                    let mut item_mut = (**item).clone();
                    stop_attacking_mut(trans, &mut item_mut, &mut mover, true).await?;
                    trans.save_item_model(&item_mut).await?;
                }
            } else {
                if let Some(ctx) = player_ctx.as_ref() {
                    trans.queue_for_session(ctx.session,
                                            Some(&format!("You try and fail to run past {}\n",
                                                          &attacker_names_str))).await?;
                }
                trans.save_item_model(&mover).await?;
                return Ok(());
            }
        }
    }
    
    if mover.is_dead {
        if !handle_resurrect(trans, &mut mover).await? {
            user_error("You couldn't be resurrected.".to_string())?;
        }
    }
    
    mover.location = format!("{}/{}", "room", new_room.code);
    mover.action_type = LocationActionType::Normal;
    mover.active_combat = None;

    trans.save_item_model(&mover).await?;

    if let Some(ctx) = player_ctx {
        look::VERB.handle(ctx, "look", "").await?;
    }

    if let Some(old_room_item) = trans.find_item_by_type_code("room", room.code).await? {
        if let Some(new_room_item) = trans.find_item_by_type_code("room", new_room.code).await? {
            announce_move(&trans, &mover, &old_room_item, &new_room_item).await?;
        }
    }

    Ok(())
}

pub struct QueueHandler;
#[async_trait]
impl QueueCommandHandler for QueueHandler {
    async fn start_command(&self, _ctx: &mut VerbContext<'_>, _command: &QueueCommand)
                           -> UResult<time::Duration> {
        Ok(time::Duration::from_secs(1))
    }

    #[allow(unreachable_patterns)]
    async fn finish_command(&self, ctx: &mut VerbContext<'_>, command: &QueueCommand)
                            -> UResult<()> {
        let direction = match command {
            QueueCommand::Movement { direction } => direction,
            _ => user_error("Unexpected command".to_owned())?
        };
        let player_item = get_player_item_or_fail(ctx).await?;
        attempt_move_immediate(ctx.trans, &player_item, direction, Some(ctx)).await?;
        Ok(())
    }
}

pub struct Verb;

#[async_trait]
impl UserVerb for Verb {
    async fn handle(self: &Self, ctx: &mut VerbContext, verb: &str, remaining: &str) -> UResult<()> {
        let dir = Direction::parse(verb).ok_or_else(|| UserError("Unknown direction".to_owned()))?;
        if remaining.trim() != "" {
            user_error("Movement commands don't take extra data at the end.".to_owned())?;
        }
        queue_command(ctx, &QueueCommand::Movement { direction: dir.clone() }).await
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
